"""Collection of schema related tags."""
from typing import Any


class Tags:
    """List of tags available."""

    BINARY_NAME_TAG = "binary_name"
    BUILD_TIMEOUT_TAG = "build_timeout"
    BUILD_TYPE_TAG = "build_type"
    COMPILER_FLAGS_TAG = "compiler_flags"
    DEADLINE_TAG = "submit_by"
    EXPECTED_OUTPUT_TAG = "expected_output"
    FOLDER_TAG = "folder"
    HOMEWORKS_TAG = "homeworks"
    INJECT_DESTINATION_TAG = "destination"
    INJECT_FOLDER_TAG = "inject_folders"
    INJECT_SOURCE_TAG = "source"
    INPUT_TAG = "input_args"
    LANGUAGE_TAG = "language"
    NAME_TAG = "name"
    OUTPUT_TYPE_TAG = "output_type"
    PIPE_TAG = "pipe_through"
    RUN_GTESTS_TAG = "run_google_tests"
    TASKS_TAG = "tasks"
    TESTS_TAG = "tests"
    TIMEOUT_TAG = "timeout"


class OutputTags:
    """Define tags for output types."""

    STRING = "string"
    NUMBER = "number"
    ALL = [STRING, NUMBER]


class BuildTags:
    """Define tags for build types."""

    CMAKE = "cmake"
    SIMPLE = "simple"
    ALL = [CMAKE, SIMPLE]


class LangTags:
    """Define tags for build types."""

    CPP = "cpp"
    BASH = "bash"
    ALL = [CPP, BASH]


class OneOf:
    """Check that an item is one of the list."""

    def __init__(self: "OneOf", some_list: list):
        """Set the list to choose from."""
        self.__items = some_list

    def __call__(self: "OneOf", item: Any):
        """Check that the list contains what is needed."""
        return item in self.__items

    def __str__(self: "OneOf"):
        """Override str for this class."""
        return "Possible values: {}".format(self.__items)



#!/usr/bin/python3
"""Test the tools."""
from __future__ import annotations

import unittest

from time import monotonic as timer
from pathlib import Path
from homework_checker.core import tools
from homework_checker.core.schema_tags import OutputTags


class TestTools(unittest.TestCase):
    """Test the checker."""

    def test_pkg_name(self: TestTools):
        """Pkg name test."""
        self.assertEqual(tools.PKG_NAME, "homework_checker")
        if tools.PROJECT_ROOT_FOLDER.exists():
            self.assertEqual(tools.PROJECT_ROOT_FOLDER.name, "homework_checker")

    def test_temp_directory_copy(self: TestTools):
        """Test that we can create and remove a temp folder."""
        folder_name = tools.PROJECT_ROOT_FOLDER
        old_temp_folder = None

        with tools.TempDirCopy(source_folder=folder_name) as tempdir:
            self.assertIn(tools.get_unique_str(str(folder_name)), str(tempdir))
            self.assertTrue(tempdir.exists())
            self.assertTrue((tempdir / tools.PKG_NAME / "core" / "tests").exists())
            with self.assertRaises(Exception):
                with tools.TempDirCopy(folder_name):
                    pass
            old_temp_folder = tempdir
        self.assertFalse(old_temp_folder.exists())

        prefix = "blah"
        with tools.TempDirCopy(source_folder=folder_name, prefix=prefix) as tempdir:
            self.assertIn(prefix, str(tempdir))

    def test_convert_to(self: TestTools):
        """Test conversion to expected type."""
        output, error = tools.convert_to(OutputTags.NUMBER, "value")
        self.assertEqual(output, None)
        self.assertEqual(error, "could not convert string to float: 'value'")

        output, error = tools.convert_to(OutputTags.STRING, 3.14)
        self.assertEqual(output, "3.14")
        self.assertEqual(error, "OK")

        output, error = tools.convert_to(OutputTags.NUMBER, "3.14")
        self.assertEqual(output, 3.14)
        self.assertEqual(error, "OK")

    def test_max_date(self: TestTools):
        """Make sure we can rely on max date."""
        self.assertEqual(tools.MAX_DATE_STR, "9999-12-31 23:59:59")

    def test_sleep_timeout(self: TestTools):
        """Test that we can break an endless loop."""
        start = timer()
        timout = 1
        cmd_result = tools.run_command("sleep 10", timeout=timout)
        self.assertFalse(cmd_result.succeeded())
        self.assertLess(timer() - start, 5)
        self.assertEqual(
            cmd_result.stderr,
            "Timeout: command 'sleep 10' ran longer than {} seconds".format(timout),
        )

    def test_git_url(self: TestTools):
        """Test that we can break an endless loop."""
        domain, user, project = tools.parse_git_url(
            "https://gitlab.ipb.uni-bonn.de/igor/some_project.git"
        )
        self.assertEqual(domain, "gitlab.ipb.uni-bonn.de")
        self.assertEqual(user, "igor")
        self.assertEqual(project, "some_project")
        domain, user, project = tools.parse_git_url(
            "git@gitlab.ipb.uni-bonn.de:igor/some_project.git"
        )
        self.assertEqual(domain, "gitlab.ipb.uni-bonn.de")
        self.assertEqual(user, "igor")
        self.assertEqual(project, "some_project")
        domain, user, project = tools.parse_git_url(
            "git@github.com:PRBonn/depth_clustering.git"
        )
        self.assertEqual(domain, "github.com")
        self.assertEqual(user, "PRBonn")
        self.assertEqual(project, "depth_clustering")

    def test_endless_loop_timeout(self: TestTools):
        """Test that we can break an endless loop."""
        path_to_file = Path(__file__).parent / "data" / "endless.cpp"
        cmd_build = "c++ -o endless -O0 {path}".format(path=str(path_to_file))
        timeout = 1.0
        cmd_result = tools.run_command(cmd_build, timeout=timeout)
        self.assertTrue(cmd_result.succeeded())
        start = timer()
        cmd_result = tools.run_command("./endless", timeout=timeout)
        self.assertFalse(cmd_result.succeeded())
        self.assertLess(timer() - start, 5)
        self.assertEqual(
            cmd_result.stderr,
            "Timeout: command './endless' ran longer than {} seconds".format(timeout),
        )



"""Write test results into a markdown file."""
from typing import Dict
from pathlib import Path

from .tools import EXPIRED_TAG
from .tools import CmdResult
from .tools import remove_number_from_name
from .checker import HomeworkResultDict

TABLE_TEMPLATE = "| {hw_name} | {task_name} | {test_name} | {result_sign} |\n"
TABLE_SEPARATOR = "|---|---|---|:---:|\n"

ERROR_TEMPLATE = """
<details><summary><b>{hw_name} | {task_name} | {test_name}</b></summary>

**`stderr`**
```apiblueprint
{stderr}
```

**`stdout`**
```
{stdout}
```

--------

</details>

"""

EXPIRED_TEMPLATE = """

<details><summary><b>{hw_name}</b></summary>

The deadline for this homework is exceeded. Errors hidden.

-------
 
</details>

"""

SEPARATOR = "--------\n"
FINISHING_NOTE = "With ❤️ from Homework Bot 🤖\n"

SUCCESS_TAG = "✅"
FAILED_TAG = "❌"


class MdWriter:
    """Write given tests results into a markdown file."""

    def __init__(self: "MdWriter"):
        """Initialize the writer."""
        self._md_table = TABLE_TEMPLATE.format(
            hw_name="Homework Name",
            task_name="Task Name",
            test_name="Test Name",
            result_sign="Result",
        )
        self._md_table += TABLE_SEPARATOR
        self._errors = ""  # Markdown part with errors.

    def update(self: "MdWriter", hw_results: Dict[str, HomeworkResultDict]):
        """Update the table of completion."""
        for hw_name, hw_dict in sorted(hw_results.items()):
            hw_name = remove_number_from_name(hw_name)
            need_hw_name = True
            expired = False
            if EXPIRED_TAG in hw_dict:
                expired = True
            for task_name, ex_dict in sorted(hw_dict.items()):
                task_name = remove_number_from_name(task_name)
                if task_name == EXPIRED_TAG:
                    # Maybe there is a better way to handle this, but I don't
                    # want to dig into this right now. We have added the
                    # EXPIRED_TAG to this dict and need to ignore it here.
                    continue
                need_task_name = True
                for test_name, test_result in sorted(ex_dict.items()):
                    test_name = remove_number_from_name(test_name)
                    result_sign = SUCCESS_TAG if test_result.succeeded() else FAILED_TAG
                    extended_hw_name = (
                        hw_name + " `[PAST DEADLINE]`" if expired else hw_name
                    )
                    self._md_table += TABLE_TEMPLATE.format(
                        hw_name=extended_hw_name if need_hw_name else "",
                        task_name=task_name if need_task_name else "",
                        test_name=test_name,
                        result_sign=result_sign,
                    )
                    self._add_error(hw_name, task_name, test_name, test_result, expired)
                    need_hw_name = False  # We only print homework name once.
                    need_task_name = False  # We only print Task name once.

    def write_md_file(self: "MdWriter", md_file_path: Path):
        """Write all the added content to the md file."""
        md_file_content = "# Test results\n"
        md_file_content += self._md_table
        if self._errors:
            md_file_content += "\n# Encountered errors\n"
            md_file_content += self._errors
        md_file_content += SEPARATOR
        md_file_content += FINISHING_NOTE
        with open(md_file_path, "w") as md_file:
            md_file.write(md_file_content)

    def _add_error(
        self: "MdWriter",
        hw_name: str,
        task_name: str,
        test_name: str,
        test_result: CmdResult,
        expired: bool,
    ):
        """Add a section of errors to the md file."""
        if test_result.succeeded():
            return
        if expired:
            self._errors += EXPIRED_TEMPLATE.format(hw_name=hw_name)
            return
        self._errors += ERROR_TEMPLATE.format(
            hw_name=hw_name,
            task_name=task_name,
            test_name=test_name,
            stderr=test_result.stderr,
            stdout=test_result.stdout,
        )



"""Handle various utility tasks."""
from __future__ import annotations
import os
import re
from os import environ
from subprocess import Popen, TimeoutExpired, CalledProcessError, CompletedProcess
from typing import Union, List, Optional, Mapping, Any, Tuple
from pathlib import Path
import tempfile
import subprocess
import logging
import datetime
import signal
import shutil
import hashlib

from .schema_tags import OutputTags

PKG_NAME = "homework_checker"
PROJECT_ROOT_FOLDER = Path(__file__).parent.parent.parent
DATE_PATTERN = "%Y-%m-%d %H:%M:%S"
MAX_DATE_STR = datetime.datetime.max.strftime(DATE_PATTERN)

EXPIRED_TAG = "expired"
NUMBER_SPLIT_TAG = "____"

log = logging.getLogger("GHC")


def get_unique_str(seed: str) -> str:
    """Generate md5 unique sting hash given init_string."""
    return hashlib.md5(seed.encode("utf-8")).hexdigest()


def add_number_to_name(number: int, name: str) -> str:
    """Add a number before a string."""
    return str(number) + NUMBER_SPLIT_TAG + name


def remove_number_from_name(name: str) -> str:
    """Add a number before a string."""
    if NUMBER_SPLIT_TAG not in name:
        return name
    return name.split(NUMBER_SPLIT_TAG)[1]


class TempDirCopy:
    """docstring for TempDirCopy"""

    def __init__(self: TempDirCopy, source_folder: Path, prefix: Optional[str] = None):
        if prefix:
            unique_temp_folder_name = "{prefix}_{name}_{unique_hash}".format(
                prefix=prefix,
                name=source_folder.name,
                unique_hash=get_unique_str(str(source_folder)),
            )
        else:
            unique_temp_folder_name = "{name}_{unique_hash}".format(
                name=source_folder.name,
                unique_hash=get_unique_str(str(source_folder)),
            )
        self.__source_folder = source_folder
        self.__temporary_folder = (
            Path(tempfile.gettempdir()) / PKG_NAME / unique_temp_folder_name
        )

    def __enter__(self: TempDirCopy) -> Path:
        if self.__temporary_folder.exists():
            raise Exception("Cannot create a temporary folder as it already exists.")
        self.__temporary_folder.mkdir(parents=True, exist_ok=True)
        try:
            shutil.copytree(
                self.__source_folder, self.__temporary_folder, dirs_exist_ok=True
            )
        except Exception as exception:
            shutil.rmtree(self.__temporary_folder)
            raise exception
        return self.__temporary_folder

    def __exit__(self: TempDirCopy, *exc_info: Any):
        shutil.rmtree(self.__temporary_folder)


def expand_if_needed(input_path: Path) -> Path:
    """Expand the path if it is not absolute."""
    if input_path.is_absolute():
        return input_path
    new_path = input_path.expanduser()
    if new_path.is_absolute():
        # This path needed user expansion. Now that the user home directory is
        # expanded this is a full absolute path.
        return new_path
    # The user could not be expanded, so we assume it is just another relative
    # path to the current working directory.
    return Path.cwd() / new_path


def convert_to(
    output_type: str, value: Any
) -> Union[Tuple[Optional[str], str], Tuple[Optional[float], str]]:
    """Convert the value to a specified type."""
    if not value:
        return None, "No value. Cannot convert {} to '{}'.".format(value, output_type)
    try:
        if output_type == OutputTags.STRING:
            return str(value).strip(), "OK"
        if output_type == OutputTags.NUMBER:
            return float(value), "OK"
    except ValueError as error:
        log.error("Exception: %s.", error)
        return None, str(error)
    return None, "Unknown output type {}. Cannot convert.".format(output_type)


def parse_git_url(git_url: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Parse the git url.

    Args:
        git_url (str): url of a git repository (https or ssh)

    Returns:
        (str, str, str): tupple of domain, user and project name parsed from url
    """
    regex = re.compile(
        r"(?:git@|https:\/\/)"  # Prefix
        r"([\w\-_\.]+)"  # Domain
        r"[:\/]"  # Separator : or /
        r"([\w\-_\.\/]+)"  # User or folders
        r"[\/]"  # Separator /
        r"([\w\-_]+)"  # Project name
        r"(?:.git)*$"
    )  # .git or nothing
    match = regex.search(git_url)
    if not match:
        return None, None, None
    domain, user, project = match.groups()
    return domain, user, project


class CmdResult:
    """A small container for command result."""

    SUCCESS = 0
    FAILURE = 13

    def __init__(
        self: CmdResult, returncode: int = None, stdout: str = None, stderr: str = None
    ):
        """Initialize either stdout of stderr."""
        self._returncode = returncode
        self._stdout = stdout
        self._stderr = stderr

    def succeeded(self: CmdResult) -> bool:
        """Check if the command succeeded."""
        if self.returncode is not None:
            return self.returncode == CmdResult.SUCCESS
        if self.stderr:
            return False
        return True

    @property
    def returncode(self: CmdResult) -> Optional[int]:
        """Get returncode."""
        return self._returncode

    @property
    def stdout(self: CmdResult) -> Optional[str]:
        """Get stdout."""
        return self._stdout

    @property
    def stderr(self: CmdResult) -> Optional[str]:
        """Get stderr."""
        return self._stderr

    @stderr.setter
    def stderr(self, value: str):
        self._returncode = None  # We can't rely on returncode anymore
        self._stderr = value

    @staticmethod
    def success() -> CmdResult:
        """Return a cmd result that is a success."""
        return CmdResult(stdout="Success!")

    def __repr__(self: CmdResult) -> str:
        """Representatin of command result."""
        stdout = self.stdout
        if not stdout:
            stdout = ""
        if self.stderr:
            return "stdout: {}, stderr: {}".format(stdout.strip(), self.stderr.strip())
        return stdout.strip()


def run_command(
    command: Union[List[str], str],
    timeout: float,
    shell: bool = True,
    cwd: Path = Path.cwd(),
    env: Optional[Mapping[str, Any]] = None,
) -> CmdResult:
    """Run a generic command in a subprocess.

    Args:
        command (str): command to run
    Returns:
        str: raw command output
    """
    try:
        startupinfo = None
        if shell and isinstance(command, list):
            command = subprocess.list2cmdline(command)
            log.debug("running command: \n%s", command)
        if env is None:
            env = environ
        process = __run_subprocess(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=shell,
            cwd=str(cwd),
            env=env,
            startupinfo=startupinfo,
            timeout=timeout,
        )
        return CmdResult(
            returncode=process.returncode,
            stdout=process.stdout.decode("utf-8"),
            stderr=process.stderr.decode("utf-8"),
        )
    except subprocess.CalledProcessError as error:
        output_text = error.output.decode("utf-8")
        log.error("command '%s' finished with code: %s", error.cmd, error.returncode)
        log.debug("command output: \n%s", output_text)
        return CmdResult(returncode=error.returncode, stderr=output_text)
    except subprocess.TimeoutExpired as error:
        output_text = "Timeout: command '{}' ran longer than {} seconds".format(
            error.cmd.strip(), error.timeout
        )
        log.error(output_text)
        return CmdResult(returncode=1, stderr=output_text)


def __run_subprocess(
    command: Union[List[str], str],
    str_input: str = None,
    timeout: float = None,
    check: bool = False,
    **kwargs
) -> subprocess.CompletedProcess:
    """Run a command as a subprocess.

    Using the guide from StackOverflow:
    https://stackoverflow.com/a/36955420/1763680
    This command has been adapted from:
    https://github.com/python/cpython/blob/3.5/Lib/subprocess.py#L352-L399

    This code does essentially the same as subprocess.run(...) but makes sure to
    kill the whole process tree which allows to use the timeout even when using
    shell=True. The reason I don't want to stop using shell=True here is the
    convenience of piping arguments from one function to another.
    """
    if str_input is not None:
        if "stdin" in kwargs:
            raise ValueError("stdin and str_input arguments may not both be used.")
        kwargs["stdin"] = subprocess.PIPE

    if timeout is None:
        timeout = 20
    with Popen(command, start_new_session=True, **kwargs) as process:
        try:
            stdout, stderr = process.communicate(str_input, timeout=timeout)
        except TimeoutExpired as timeout_error:
            # Kill the whole group of processes.
            os.killpg(process.pid, signal.SIGINT)
            stdout, stderr = process.communicate()
            raise TimeoutExpired(
                process.args, timeout, output=stdout, stderr=stderr
            ) from timeout_error
        retcode = process.poll()
        if retcode is None:
            retcode = 1
        if check and retcode:
            raise CalledProcessError(
                retcode, process.args, output=stdout, stderr=stderr
            )
    return CompletedProcess(process.args, retcode, stdout, stderr)



#!/usr/bin/python3
"""Script to check this homework."""
import argparse
import logging
from pathlib import Path

from homework_checker.core.checker import Checker
from homework_checker.core.md_writer import MdWriter
from homework_checker.core.tools import expand_if_needed

logging.basicConfig()
log = logging.getLogger("GHC")
log.setLevel(logging.INFO)


def main():
    """Run this script."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--verbose", help="Make the output verbose.", action="store_true"
    )
    parser.add_argument(
        "-i",
        "--input",
        help="An input *.yml file with the job definition.",
        required=True,
    )
    parser.add_argument(
        "-o", "--output", help="An output *.md file with the results.", required=True
    )
    args = parser.parse_args()
    if args.verbose:
        log.setLevel(logging.DEBUG)
        log.debug("Enable DEBUG logging.")
    input_file = expand_if_needed(Path(args.input))
    log.debug('Reading from file "%s"', input_file)
    checker = Checker(input_file)
    results = checker.check_all_homeworks()
    md_writer = MdWriter()
    md_writer.update(results)
    output_file = expand_if_needed(Path(args.output))
    log.debug('Writing to file "%s"', output_file)
    md_writer.write_md_file(output_file)


if __name__ == "__main__":
    main()



#!/usr/bin/python3
"""Test the checker."""
from __future__ import annotations

import unittest

from pathlib import Path
from typing import Dict

from homework_checker.core.checker import Checker
from homework_checker.core.md_writer import MdWriter
from homework_checker.core import tools


class TestChecker(unittest.TestCase):
    """Test the checker."""

    @staticmethod
    def __sanitize_results(results: Dict[str, Dict[str, tools.CmdResult]]):
        sanitized_results = {}
        for key, value in results.items():
            sanitized_value = {}
            for inner_key, inner_value in value.items():
                sanitized_value[tools.remove_number_from_name(inner_key)] = inner_value
            sanitized_results[tools.remove_number_from_name(key)] = sanitized_value
        return sanitized_results

    def test_everything(self: TestChecker):
        """Check all homeworks and Tasks."""
        path_to_job = (
            tools.PROJECT_ROOT_FOLDER
            / "homework_checker"
            / "core"
            / "tests"
            / "data"
            / "homework"
            / "example_job.yml"
        )
        checker = Checker(path_to_job)
        results = checker.check_all_homeworks()
        results = TestChecker.__sanitize_results(results)
        self.assertEqual(len(results), 3)
        self.assertNotIn("Non existing homework", results)

        self.assertIn("Homework where things go wrong", results)
        self.assertIn(tools.EXPIRED_TAG, results["Homework where things go wrong"])

        self.assertEqual(
            len(results["Homework where things go wrong"]),
            3,
            "Wrong results: {}".format(results["Homework where things go wrong"]),
        )
        self.assertIn("Return number task", results["Homework where things go wrong"])
        self.assertIn("While loop task", results["Homework where things go wrong"])
        self.assertNotIn("Non existing task", results["Homework where things go wrong"])

        writer = MdWriter()
        writer.update(results)
        writer.write_md_file(Path("results.md"))



#!/usr/bin/python3
"""A script to parse an input git url and get it's wiki counterpart.

Attributes:
    WIKI_REPO_MASK (str): mask of wiki git repo
    REPO_MASK (str): mask of the git repo
"""
import sys

from homework_checker.core.tools import parse_git_url

WIKI_REPO_MASK = "git@{domain}:{user}/{project}.wiki.git"
REPO_MASK = "git@{domain}:{user}/{project}.git"


def main():
    """Print the name of the repo."""
    if len(sys.argv) < 3:
        print("ERROR: must be supplied with a git url and type [wiki|code]")
        print(
            "[Example]: {binary} {repo} {type}".format(
                binary="python3 print_repo_name.py",
                repo="git@gitlab.igg.uni-bonn.de:igor/some_project.git",
                type="wiki",
            )
        )
        print(
            "[Example]: {binary} {repo} {type}".format(
                binary="python3 print_repo_name.py",
                repo="https://gitlab.ipb.uni-bonn.de/igor/some_project.git",
                type="code",
            )
        )
        sys.exit(1)
    if len(sys.argv) == 3:
        repo = sys.argv[1]
        domain, user, project = parse_git_url(repo)
        repo_type = sys.argv[2]
        if repo_type == "wiki":
            print(WIKI_REPO_MASK.format(domain=domain, user=user, project=project))
        elif repo_type == "code":
            print(REPO_MASK.format(domain=domain, user=user, project=project))
        else:
            print('ERROR: type "{}" is not "wiki" or "code"'.format(repo_type))


if __name__ == "__main__":
    main()



"""Check the homework."""

import logging
from datetime import datetime
from typing import Dict
from pathlib import Path

from . import tools
from .schema_manager import SchemaManager
from .schema_tags import Tags
from .tasks import Task


log = logging.getLogger("GHC")

HomeworkResultDict = Dict[str, Dict[str, tools.CmdResult]]


class Checker:
    """Check homework."""

    TESTS_TAG = "tests"
    NUMBERING_TAG = "____"

    def __init__(self: "Checker", job_file_path: Path):
        """Initialize the checker from file."""
        self._job_file_path = tools.expand_if_needed(job_file_path)
        schema_manager = SchemaManager(self._job_file_path)
        self._base_node = schema_manager.validated_yaml
        self._checked_code_folder = tools.expand_if_needed(
            Path(self._base_node[Tags.FOLDER_TAG])
        )
        log.debug("self._checked_code_folder: %s", self._checked_code_folder)

    def check_homework(self: "Checker", homework_node: dict) -> HomeworkResultDict:
        """Run over all Tasks in a single homework."""
        results: HomeworkResultDict = {}
        current_folder = Path(self._checked_code_folder, homework_node[Tags.FOLDER_TAG])
        log.debug("current_folder: %s", current_folder)
        if not current_folder.exists():
            log.warning("Folder '%s' does not exist. Skiping.", current_folder)
            return results
        deadline_str = homework_node[Tags.DEADLINE_TAG]
        deadline_datetime = datetime.strptime(deadline_str, tools.DATE_PATTERN)
        if datetime.now() > deadline_datetime:
            results[tools.EXPIRED_TAG] = {}
        for idx, task_node in enumerate(homework_node[Tags.TASKS_TAG]):
            task = Task.from_yaml_node(
                task_node=task_node,
                student_hw_folder=current_folder,
                job_file=self._job_file_path,
            )
            if not task:
                continue
            results[tools.add_number_to_name(idx, task.name)] = task.check()
        return results

    def check_all_homeworks(self: "Checker") -> Dict[str, HomeworkResultDict]:
        """Run over all Tasks in all homeworks."""
        results: Dict[str, HomeworkResultDict] = {}
        for idx, homework_node in enumerate(self._base_node[Tags.HOMEWORKS_TAG]):
            hw_name = tools.add_number_to_name(idx, homework_node[Tags.NAME_TAG])
            current_homework_results = self.check_homework(homework_node)
            if current_homework_results:
                results[hw_name] = current_homework_results
        return results



"""Manage creation of schema."""
from __future__ import annotations
import sys
import logging
import operator
from pathlib import Path
from typing import Union, List, Dict, Any
from ruamel.yaml import YAML  # type: ignore[import]
from ruamel.yaml.comments import CommentedMap, CommentedSeq  # type: ignore[import]
from schema import Schema, SchemaError, Or, Optional  # type: ignore[import]

from .tools import MAX_DATE_STR
from .schema_tags import Tags, OutputTags, BuildTags, LangTags

log = logging.getLogger("GHC")

SCHEMA_FILE = Path(__file__).parent.parent / "schema" / "schema.yml"


class SchemaManager:
    """Manage schema creation."""

    GenericListType = List[Union[Any, Dict[str, Any], List[Any]]]
    GenericDictType = Dict[str, Union[Any, Dict[str, Any], List[Any]]]

    def __init__(self: SchemaManager, file_name: Path):
        """Create a schema for my tests."""

        injection_schema = {
            Tags.INJECT_SOURCE_TAG: str,
            Tags.INJECT_DESTINATION_TAG: str,
        }

        test_schema = {
            Tags.NAME_TAG: str,
            Optional(Tags.INPUT_TAG): str,
            Optional(Tags.INJECT_FOLDER_TAG): [injection_schema],
            Optional(Tags.RUN_GTESTS_TAG, default=False): bool,
            Optional(Tags.EXPECTED_OUTPUT_TAG): Or(str, float, int),
            Optional(Tags.TIMEOUT_TAG, default=60): float,
        }

        task_schema = {
            Tags.NAME_TAG: str,
            Tags.LANGUAGE_TAG: Or(LangTags.CPP, LangTags.BASH),
            Tags.FOLDER_TAG: str,
            Optional(Tags.OUTPUT_TYPE_TAG, default=OutputTags.STRING): Or(
                OutputTags.STRING, OutputTags.NUMBER
            ),
            Optional(Tags.COMPILER_FLAGS_TAG, default="-Wall"): str,
            Optional(Tags.BINARY_NAME_TAG, default="main"): str,
            Optional(Tags.PIPE_TAG, default=""): str,
            Optional(Tags.BUILD_TYPE_TAG, default=BuildTags.CMAKE): Or(
                BuildTags.CMAKE, BuildTags.SIMPLE
            ),
            Optional(Tags.BUILD_TIMEOUT_TAG, default=60): float,
            Optional(Tags.TESTS_TAG): [test_schema],
        }

        homework_schema = {
            Tags.NAME_TAG: str,
            Tags.FOLDER_TAG: str,
            Optional(Tags.DEADLINE_TAG, default=MAX_DATE_STR): str,
            Tags.TASKS_TAG: [task_schema],
        }

        self.__schema = Schema(
            {
                Tags.FOLDER_TAG: str,
                Tags.HOMEWORKS_TAG: [homework_schema],
            }
        )
        yaml = YAML()
        # big enough value to prevent wrapping
        yaml.width = 4096  # type: ignore[assignment]
        yaml.explicit_start = True  # type: ignore[assignment]
        yaml.indent(mapping=2, sequence=4, offset=2)
        with open(file_name, "r") as stream:
            contents = SchemaManager.__to_simple_dict(yaml.load(stream))
            try:
                self.__validated_yaml = self.__schema.validate(contents)
            except SchemaError as exc:
                sys.exit(exc.code)
        # Write the schema every time we run this code while developing. We
        # don't want to run this when the package is installed as this we won't
        # have the permission. This is intended to keep the schema file up to
        # date when we add new stuff to it.
        try:
            with open(SCHEMA_FILE, "w") as outfile:
                str_dict = SchemaManager.__sanitize_value(self.__schema._schema)
                yaml.dump(str_dict, outfile)
        except OSError:
            log.debug("Cannot write schema file. We only use this while developing.")

    @staticmethod
    def __to_simple_list(
        commented_seq: List[Union[Any, CommentedSeq, CommentedMap]]
    ) -> SchemaManager.GenericListType:
        simple_list: SchemaManager.GenericListType = []
        for value in commented_seq:
            if isinstance(value, CommentedSeq):
                simple_list.append(SchemaManager.__to_simple_list(value))
            elif isinstance(value, CommentedMap):
                simple_list.append(SchemaManager.__to_simple_dict(value))
            else:
                simple_list.append(value)
        return simple_list

    @staticmethod
    def __to_simple_dict(
        commented_map: Dict[str, Union[Any, CommentedSeq, CommentedMap]],
    ) -> SchemaManager.GenericDictType:
        simple_dict: SchemaManager.GenericDictType = {}
        for key, value in commented_map.items():
            if isinstance(value, CommentedMap):
                simple_dict[key] = SchemaManager.__to_simple_dict(value)
            elif isinstance(value, CommentedSeq):
                simple_dict[key] = SchemaManager.__to_simple_list(value)
            else:
                simple_dict[key] = value
        return simple_dict

    @property
    def validated_yaml(self) -> dict:
        """Return validated yaml."""
        return self.__validated_yaml

    @property
    def schema(self):
        """Return schema."""
        return self.__schema

    # pylint: disable=R0911
    # This method needs this many returns.
    @staticmethod
    def __sanitize_value(input_var: Union[dict, Optional, Or, Any]):
        """Use the schema and create an example file from it."""
        # pylint: disable=W0212
        # This seems to be the only way to get to schema value.
        if isinstance(input_var, dict):
            new_dict = {}
            for key, val in input_var.items():
                new_dict[
                    SchemaManager.__sanitize_value(key)
                ] = SchemaManager.__sanitize_value(val)
            return CommentedMap(sorted(new_dict.items(), key=operator.itemgetter(0)))
        if isinstance(input_var, list):
            new_list = []
            for val in input_var:
                new_list.append(SchemaManager.__sanitize_value(val))
            return new_list
        if isinstance(input_var, Optional):
            if input_var._schema == Tags.DEADLINE_TAG:
                return SchemaManager.__sanitize_value(input_var._schema)
            return "~[optional]~ " + SchemaManager.__sanitize_value(input_var._schema)
        if isinstance(input_var, Or):
            return "Any of " + str(
                [SchemaManager.__sanitize_value(s) for s in input_var._args]
            )
        if input_var is str:
            return "String value"
        if input_var is float:
            return "Float value"
        if input_var is int:
            return "Int value"
        if input_var is bool:
            return "Boolean value"
        return str(input_var)



"""Different types of Tasks."""
from __future__ import annotations

import logging
import abc
from pathlib import Path
from typing import Optional, Dict, List, Tuple
from shutil import copytree, rmtree

from . import tools
from .schema_tags import Tags, LangTags, BuildTags


log = logging.getLogger("GHC")


OUTPUT_MISMATCH_MESSAGE = """Given input: '{input}'
Your output '{actual}'
Expected output: '{expected}'"""

BUILD_SUCCESS_TAG = "Build succeeded"
STYLE_ERROR_TAG = "Style errors"

TOTAL_ERRORS_FOUND_TAG = "Total errors found"


class Task:
    """Define an abstract Task."""

    BACKUP_FOLDER = ".backup"

    ResultDictType = Dict[str, tools.CmdResult]

    class Injection:
        """A small class that encapsulates an injection source and destination."""

        def __init__(self: Task.Injection, source: Path, destination: Path):
            self._source = source
            self._destination = destination

        @property
        def source(self) -> Path:
            """Get the source folder."""
            return self._source

        @property
        def destination(self) -> Path:
            """Get the destination folder."""
            return self._destination

    @staticmethod
    def from_yaml_node(
        task_node: dict, student_hw_folder: Path, job_file: Path
    ) -> Optional[Task]:
        """Create an Task appropriate for the language."""
        student_task_folder = student_hw_folder / task_node[Tags.FOLDER_TAG]
        if not student_task_folder.exists():
            log.warning("Folder '%s' does not exist. Skipping.", student_task_folder)
            return None
        language_tag = task_node[Tags.LANGUAGE_TAG]
        if language_tag == LangTags.CPP:
            return CppTask(task_node, student_task_folder, job_file)
        if language_tag == LangTags.BASH:
            return BashTask(task_node, student_task_folder, job_file)
        log.error("Unknown Task language.")
        return None

    def __init__(
        self: Task, task_node: dict, student_task_folder: Path, job_file: Path
    ):
        """Initialize a generic Task."""
        self.name = task_node[Tags.NAME_TAG]
        self._job_root_folder = job_file.parent
        self._output_type = task_node[Tags.OUTPUT_TYPE_TAG]
        self._student_task_folder = student_task_folder
        self._binary_name = task_node[Tags.BINARY_NAME_TAG]
        self._pipe_through = task_node[Tags.PIPE_TAG]
        self._build_timeout = task_node[Tags.BUILD_TIMEOUT_TAG]
        if Tags.TESTS_TAG in task_node:
            self._test_nodes = task_node[Tags.TESTS_TAG]
        else:
            self._test_nodes = []  # Sometimes we don't have tests.
        self.__test_counter = 0

    def __with_number_prefix(self: Task, test_name: str) -> str:
        """Get the test name with number."""
        self.__test_counter += 1
        return tools.add_number_to_name(self.__test_counter, test_name)

    def check(self: Task) -> Task.ResultDictType:
        """Iterate over the tests and check them."""
        # Generate empty results.
        results: Task.ResultDictType = {}

        def run_all_tests(
            test_node: dict, executable_folder: Path
        ) -> Task.ResultDictType:
            """Run all tests in the task."""
            results: Task.ResultDictType = {}
            if Tags.INJECT_FOLDER_TAG not in test_node:
                # There is no need to rebuild the code. We can just run our tests.
                test_result = self._run_test(test_node, executable_folder)
                results[
                    self.__with_number_prefix(test_node[Tags.NAME_TAG])
                ] = test_result
                return results
            # There are folders to inject, so we will have to rebuild with the newly
            # injected folders. We do it in a new temp folder.
            with tools.TempDirCopy(
                source_folder=self._student_task_folder, prefix="build_injected"
            ) as code_folder:
                folders_to_inject = self.__get_folders_to_inject(
                    node=test_node, destination_root=code_folder
                )
                Task.__inject_folders(folders_to_inject)
                build_result, build_folder = self._build_if_needed(code_folder)
                if build_result and not build_result.succeeded():
                    raise Exception("Build with inserted folders must ALWAYS succeed!")
                test_result = self._run_test(
                    test_node=test_node, executable_folder=build_folder
                )
                results[
                    self.__with_number_prefix(test_node[Tags.NAME_TAG])
                ] = test_result
            return results

        with tools.TempDirCopy(self._student_task_folder) as code_folder:
            # Build the source if this is needed.
            build_result, build_folder = self._build_if_needed(code_folder)
            if build_result:
                results[self.__with_number_prefix(BUILD_SUCCESS_TAG)] = build_result
                if not build_result.succeeded():
                    # The build has failed, so no further testing needed.
                    return results
            # The build is either not needed or succeeded. Continue testing.
            for test_node in self._test_nodes:
                results.update(
                    run_all_tests(test_node=test_node, executable_folder=build_folder)
                )

        style_errors = self._code_style_errors()
        if style_errors:
            results[STYLE_ERROR_TAG] = style_errors
        return results

    def __get_folders_to_inject(
        self: Task, node: dict, destination_root: Path
    ) -> List[Injection]:
        folders_to_inject = []
        if Tags.INJECT_FOLDER_TAG in node:
            # Inject all needed folders.
            for injection in node[Tags.INJECT_FOLDER_TAG]:
                source_folder = (
                    self._job_root_folder / injection[Tags.INJECT_SOURCE_TAG]
                )
                destination_folder = (
                    destination_root / injection[Tags.INJECT_DESTINATION_TAG]
                )
                folders_to_inject.append(
                    Task.Injection(source=source_folder, destination=destination_folder)
                )
        return folders_to_inject

    @staticmethod
    def __inject_folders(folders_to_inject: List[Task.Injection]):
        """Inject all folders overwriting existing folders in case of conflict."""
        for injection in folders_to_inject:
            if injection.destination.exists():
                rmtree(injection.destination)
            copytree(injection.source, injection.destination)

    @abc.abstractmethod
    def _run_test(self: Task, test_node: dict, executable_folder: Path):
        return None

    @abc.abstractmethod
    def _build_if_needed(self: Task, code_folder: Path):
        return None, code_folder

    @abc.abstractmethod
    def _code_style_errors(self: Task):
        return None


class CppTask(Task):
    """Define a C++ Task."""

    CMAKE_BUILD_CMD = "cmake .. && make -j2"
    REMAKE_AND_TEST = "make clean && rm -r * && cmake .. && make -j2 && ctest -VV"
    BUILD_CMD_SIMPLE = "clang++ -std=c++14 -o {binary} {compiler_flags} {binary}.cpp"

    def __init__(self: CppTask, task_node: dict, root_folder: Path, job_file: Path):
        """Initialize the C++ Task."""
        super().__init__(task_node, root_folder, job_file)
        self._compiler_flags = task_node[Tags.COMPILER_FLAGS_TAG]
        self._build_type = task_node[Tags.BUILD_TYPE_TAG]

    def _build_if_needed(
        self: CppTask, code_folder: Path
    ) -> Tuple[tools.CmdResult, Path]:
        if self._build_type == BuildTags.CMAKE:
            build_folder = code_folder / "build"
            build_folder.mkdir(parents=True, exist_ok=True)
            return (
                tools.run_command(
                    CppTask.CMAKE_BUILD_CMD,
                    cwd=build_folder,
                    timeout=self._build_timeout,
                ),
                build_folder,
            )
        return (
            tools.run_command(
                CppTask.BUILD_CMD_SIMPLE.format(
                    binary=self._binary_name, compiler_flags=self._compiler_flags
                ),
                cwd=code_folder,
                timeout=self._build_timeout,
            ),
            code_folder,
        )

    def _code_style_errors(self: CppTask) -> Optional[tools.CmdResult]:
        """Check if code conforms to Google Style."""
        command = (
            "cpplint --counting=detailed "
            + "--filter=-legal,-readability/todo,"
            + "-build/include_order,-runtime/threadsafe_fn,"
            + "-runtime/arrays"
            + ' $( find . -name "*.h" -o -name "*.cpp" | grep -vE "^./build/" )'
        )
        result = tools.run_command(
            command,
            cwd=self._student_task_folder,
            timeout=self._build_timeout,
        )
        if result.stderr and TOTAL_ERRORS_FOUND_TAG in result.stderr:
            return result
        if result.stdout and TOTAL_ERRORS_FOUND_TAG in result.stdout:
            return result
        return None

    def _run_test(self: CppTask, test_node: dict, executable_folder: Path):
        if test_node[Tags.RUN_GTESTS_TAG]:
            return tools.run_command(
                CppTask.REMAKE_AND_TEST,
                cwd=executable_folder,
                timeout=test_node[Tags.TIMEOUT_TAG],
            )
        input_str = ""
        if Tags.INPUT_TAG in test_node:
            input_str = test_node[Tags.INPUT_TAG]
        run_cmd = "./{binary_name} {args}".format(
            binary_name=self._binary_name, args=input_str
        )
        if self._pipe_through:
            run_cmd += " " + self._pipe_through
        run_result = tools.run_command(
            run_cmd, cwd=executable_folder, timeout=test_node[Tags.TIMEOUT_TAG]
        )
        if not run_result.succeeded():
            return run_result
        # TODO(igor): do I need explicit error here?
        our_output, error = tools.convert_to(self._output_type, run_result.stdout)
        if not our_output:
            # Conversion has failed.
            run_result.stderr = error
            return run_result
        expected_output, error = tools.convert_to(
            self._output_type, test_node[Tags.EXPECTED_OUTPUT_TAG]
        )
        if our_output != expected_output:
            run_result.stderr = OUTPUT_MISMATCH_MESSAGE.format(
                actual=our_output, input=input_str, expected=expected_output
            )
        return run_result


class BashTask(Task):
    """Define a Bash Task."""

    RUN_CMD = "sh {binary_name}.sh {args}"

    def __init__(self: BashTask, task_node: dict, root_folder: Path, job_file: Path):
        """Initialize the Task."""
        super().__init__(task_node, root_folder, job_file)

    def _build_if_needed(self: BashTask, code_folder: Path):
        return None, code_folder  # There is nothing to build in Bash.

    def _code_style_errors(self: BashTask):
        return None

    def _run_test(
        self: BashTask, test_node: dict, executable_folder: Path
    ) -> tools.CmdResult:
        input_str = ""
        if Tags.INPUT_TAG in test_node:
            input_str = test_node[Tags.INPUT_TAG]
        run_cmd = BashTask.RUN_CMD.format(binary_name=self._binary_name, args=input_str)
        if self._pipe_through:
            run_cmd += " " + self._pipe_through
        run_result = tools.run_command(
            run_cmd, cwd=executable_folder, timeout=test_node[Tags.TIMEOUT_TAG]
        )
        if not run_result.succeeded():
            return run_result
        our_output, error = tools.convert_to(self._output_type, run_result.stdout)
        if not our_output:
            # Conversion has failed.
            run_result.stderr = error
            return run_result
        expected_output, error = tools.convert_to(
            self._output_type, test_node[Tags.EXPECTED_OUTPUT_TAG]
        )
        if our_output != expected_output:
            run_result.stderr = OUTPUT_MISMATCH_MESSAGE.format(
                actual=our_output, input=input_str, expected=expected_output
            )
        return run_result



#!/usr/bin/python3
"""Test the checker."""

import unittest
from pathlib import Path
from typing import Tuple

from homework_checker.core.tasks import (
    Task,
    BUILD_SUCCESS_TAG,
    STYLE_ERROR_TAG,
)
from homework_checker.core.schema_tags import Tags
from homework_checker.core.schema_manager import SchemaManager
from homework_checker.core import tools


class TestTask(unittest.TestCase):
    """Test the checker."""

    def setUp(self: "TestTask"):
        # pylint: disable=C0103
        self.maxDiff = None
        self.job_file_path = (
            tools.PROJECT_ROOT_FOLDER
            / "homework_checker"
            / "core"
            / "tests"
            / "data"
            / "homework"
            / "example_job.yml"
        )
        schema_manager = SchemaManager(self.job_file_path)
        base_node = schema_manager.validated_yaml
        self.assertIn(Tags.HOMEWORKS_TAG, base_node)
        self.homework_nodes = base_node[Tags.HOMEWORKS_TAG]
        self.assertEqual(len(self.homework_nodes), 4)
        self.checked_code_folder = tools.expand_if_needed(
            Path(base_node[Tags.FOLDER_TAG])
        )

    def __get_homework_name_and_task(
        self: "TestTask", homework_index: int, task_index: int
    ) -> Tuple[str, Task]:
        homework_node = self.homework_nodes[homework_index]
        homework_folder = self.checked_code_folder / homework_node[Tags.FOLDER_TAG]
        task_node = homework_node[Tags.TASKS_TAG][task_index]
        task = Task.from_yaml_node(
            task_node=task_node,
            student_hw_folder=homework_folder,
            job_file=self.job_file_path,
        )
        if task is None:
            # We cannot use assertIsNotNone as this confuses mypy.
            self.fail("Task cannot be None")
        return homework_node[Tags.NAME_TAG], task

    @staticmethod
    def __sanitize_results(results: dict) -> dict:
        """Sanitize the outputs of the tasks."""
        sanitized_results = {}
        for key, value in results.items():
            sanitized_results[tools.remove_number_from_name(key)] = value
        return sanitized_results

    def test_check_simple_cpp_io_task(self: "TestTask"):
        """Check that we can build and run cpp code and get some output."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=0, task_index=0
        )
        self.assertEqual(homework_name, "Sample homework")
        self.assertEqual(task.name, "Simple cpp tasks")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 3
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs + expected_number_of_test_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results[BUILD_SUCCESS_TAG].succeeded())
        self.assertEqual(results["String output test"].stderr, "")
        self.assertTrue(results["String output test"].succeeded())
        self.assertTrue(results["Input output test"].succeeded())
        self.assertFalse(results["Wrong output fail"].succeeded())

    def test_check_build_failure_task(self: "TestTask"):
        """Check that the task fails if code does not build."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=0, task_index=1
        )
        self.assertEqual(homework_name, "Sample homework")
        self.assertEqual(task.name, "Build failure task")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs + expected_number_of_test_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertFalse(results[BUILD_SUCCESS_TAG].succeeded())

    def test_check_cmake_arithmetics_task(self: "TestTask"):
        """Check a simple cmake build on arithmetics example."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=0, task_index=2
        )
        self.assertEqual(homework_name, "Sample homework")
        self.assertEqual(task.name, "CMake build arithmetics task")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 2
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results[BUILD_SUCCESS_TAG].succeeded())
        self.assertTrue(results["Test integer arithmetics"].succeeded())
        self.assertFalse(results["Test float arithmetics"].succeeded())

    def test_check_bash_task(self: "TestTask"):
        """Check a simple cmake build on arithmetics example."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=0, task_index=3
        )
        self.assertEqual(homework_name, "Sample homework")
        self.assertEqual(task.name, "Bash task")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 0
        expected_number_of_test_outputs = 2
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        print(results)
        self.assertTrue(results["Test output"].succeeded())
        self.assertFalse(results["Test wrong output"].succeeded())

    def test_check_wrong_output_type(self: "TestTask"):
        """Check that we detect a wrong output type."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=1, task_index=0
        )
        self.assertEqual(homework_name, "Homework where things go wrong")
        self.assertEqual(task.name, "Return number task")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 1
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results[BUILD_SUCCESS_TAG].succeeded())
        self.assertFalse(results["Wrong output format"].succeeded())
        self.assertEqual(
            results["Wrong output format"].stderr,
            "could not convert string to float: 'hello world\\n'",
        )

    def test_timing_out_task(self: "TestTask"):
        """Check that we can deal with infinitely running code."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=1, task_index=1
        )
        self.assertEqual(homework_name, "Homework where things go wrong")
        self.assertEqual(task.name, "While loop task")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 1
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results[BUILD_SUCCESS_TAG].succeeded())
        self.assertFalse(results["Test timeout"].succeeded())
        if not results["Test timeout"].stderr:
            self.fail()
        self.assertRegex(
            results["Test timeout"].stderr,
            r"Timeout: command './main' ran longer than .* seconds",
        )

    def test_google_tests_task(self: "TestTask"):
        """Check that we can run google test."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=2, task_index=0
        )
        self.assertEqual(homework_name, "Homework with injections")
        self.assertEqual(task.name, "Google Tests")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 1
        expected_number_of_test_outputs = 3
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results[BUILD_SUCCESS_TAG].succeeded())
        self.assertTrue(results["Just build"].succeeded())
        self.assertTrue(results["Inject pass"].succeeded())
        self.assertFalse(results["Inject fail"].succeeded())

    def test_bash_task_with_injections(self: "TestTask"):
        """Check that we can inject folders into bash tasks too."""

        homework_name, task = self.__get_homework_name_and_task(
            homework_index=2, task_index=1
        )
        self.assertEqual(homework_name, "Homework with injections")
        self.assertEqual(task.name, "Bash with many folders")
        results = task.check()
        results = TestTask.__sanitize_results(results)
        expected_number_of_build_outputs = 0
        expected_number_of_test_outputs = 1
        expected_number_of_code_style_outputs = 0
        self.assertEqual(
            len(results),
            expected_number_of_build_outputs
            + expected_number_of_test_outputs
            + expected_number_of_code_style_outputs,
            "Wrong results: {}".format(results),
        )
        self.assertTrue(results["ls"].succeeded())



import os
import sys
from stat import ST_MODE
from distutils import log
from setuptools import setup
from setuptools import find_packages
from setuptools.command.install import install

VERSION_STRING = "1.0.3"

PACKAGE_NAME = "homework_checker"

# Setup installation dependencies
INSTALL_REQUIRES = [
    "ruamel.yaml",
    "schema",
    "setuptools",
    "cpplint",
    "datetime",
]

if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
    INSTALL_REQUIRES.append("argparse")


class PermissiveInstall(install):
    """A class for permissive install."""

    def run(self):
        """Run the install procedure."""
        install.run(self)
        if os.name == "posix":
            for file in self.get_outputs():
                # all installed files should be readable for anybody
                mode = ((os.stat(file)[ST_MODE]) | 0o444) & 0o7777
                log.info("changing permissions of %s to %o" % (file, mode))
                os.chmod(file, mode)


GITHUB_URL = "https://github.com/cpp-for-yourself/{}".format(PACKAGE_NAME)

setup(
    name=PACKAGE_NAME,
    version=VERSION_STRING,
    install_requires=INSTALL_REQUIRES,
    setup_requires=["nose>=1.0"],
    author="Igor Bogoslavskyi",
    author_email="igor.bogoslavskyi@gmail.com",
    maintainer="Igor Bogoslavskyi",
    maintainer_email="igor.bogoslavskyi@gmail.com",
    keywords=["homework-checker"],
    license="Apache 2.0",
    url=GITHUB_URL,
    download_url=GITHUB_URL + "/tarball/" + VERSION_STRING,
    classifiers=[
        "Environment :: Console",
        "Programming Language :: Python",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
    ],
    description="""A generic homework checker.""",
    long_description=open("README.md").read(),
    long_description_content_type='text/markdown',
    test_suite="tests",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "check_homework = homework_checker.check_homework:main",
            "print_repo_name = homework_checker.print_repo_name:main",
        ],
    },
    cmdclass={"install": PermissiveInstall},
)
